/*
 * The MIT License (MIT)
 *
 * Copyright © 2023 xrv <xrv@live.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.tzld.crawler.etl.mq;

import com.alibaba.fastjson.JSONObject;
import com.aliyun.mq.http.MQClient;
import com.aliyun.mq.http.MQConsumer;
import com.aliyun.mq.http.model.Message;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.tzld.crawler.etl.model.vo.CrawlerEtlParam;
import com.tzld.crawler.etl.service.EtlService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * @author ehlxr
 * @since 2023-06-09 15:24.
 */
@Component
public class EtlMQConsumer {
    private static final Logger log = LoggerFactory.getLogger(EtlMQConsumer.class);
    private final EtlService etlService;
    @Value("${rocketmq.accessKey:}")
    private String accessKey;
    @Value("${rocketmq.secretKey:}")
    private String secretKey;
    @Value("${rocketmq.httpEndpoint:}")
    private String httpEndpoint;
    @Value("${rocketmq.instanceId:}")
    private String instanceId;
    @Value("${rocketmq.crawler.etl.topic:}")
    private String topic;
    @Value("${rocketmq.crawler.etl.groupid:}")
    private String groupId;

    @ApolloJsonValue("${crawler.etl.priority:{}}")
    private Map<String, Integer> priorityMap;

    private MQConsumer consumer;
    private ThreadPoolExecutor priorityPool;

    public EtlMQConsumer(EtlService etlService) {
        this.etlService = etlService;
    }

    @PostConstruct
    public void init() {
        MQClient mqClient = new MQClient(httpEndpoint, accessKey, secretKey);
        priorityPool = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors(),
                Runtime.getRuntime().availableProcessors() * 2, 60L,
                TimeUnit.SECONDS, new PriorityBlockingQueue<>(),
                new ThreadFactoryBuilder().setNameFormat("priority-etl-pool-%d").build());
        consumer = mqClient.getConsumer(instanceId, topic, groupId, null);

        new Thread(this::doConsume).start();
    }

    private void doConsume() {
        do {
            List<String> handles = new ArrayList<>();
            try {
                List<Message> messages = consumer.consumeMessage(1, 10);
                if (messages == null || messages.isEmpty()) {
                    log.info("No new message, continue");
                    continue;
                }

                messages.forEach(message -> {
                    logPoolInfo();
                    if (priorityPool.getQueue().size() < 10) {
                        log.info("Receive message: {} from topic: {}, group: {}", message, topic, groupId);
                        CrawlerEtlParam param = JSONObject.parseObject(message.getMessageBodyString(), CrawlerEtlParam.class);
                        param.setMessageId(message.getMessageId());

                        priorityPool.execute(new RunnablePriority(param));
                        handles.add(message.getReceiptHandle());
                    } else {
                        log.warn("Priority pool queue size > {}! waiting for execute task.", priorityPool.getQueue().size());
                    }
                });
            } catch (Throwable e) {
                log.error("Consume message from topic: {}, group: {} error", topic, groupId, e);
            }

            if (!handles.isEmpty()) {
                consumer.ackMessage(handles);
            }
        } while (true);
    }

    private void logPoolInfo() {
        log.info("Thread pool info TaskCount {}, CompletedTaskCount {}, ActiveCount {}, Queue size {}, Queue remainingCapacity {}",
                priorityPool.getTaskCount(), priorityPool.getCompletedTaskCount(), priorityPool.getActiveCount(),
                priorityPool.getQueue().size(), priorityPool.getQueue().remainingCapacity());
    }

    class RunnablePriority implements Runnable, Comparable<RunnablePriority> {
        private final CrawlerEtlParam param;

        public RunnablePriority(CrawlerEtlParam param) {
            this.param = param;
        }

        public Integer getPriority() {
            return priorityMap.getOrDefault(param.getPlatform(), -1);
        }

        /**
         * 优先级比较
         */
        @Override
        public int compareTo(RunnablePriority o) {
            if (this.getPriority() < o.getPriority()) {
                return 1;
            } else if (this.getPriority() > o.getPriority()) {
                return -1;
            }
            return 0;
        }

        @Override
        public void run() {
            try {
                logPoolInfo();
                log.info("deal: {} priority: {}", param, getPriority());
                etlService.deal(param);
                logPoolInfo();
            } catch (Exception e) {
                log.error("deal {} error.", param, e);
            }
        }

        @Override
        public String toString() {
            return param.getPlatform() + ":" + param.getOutVideoId();
        }
    }
}
